function varargout = nlpsol_default_in(varargin)
    %NLPSOL_DEFAULT_IN [INTERNAL] 
    %
    %  [double] = NLPSOL_DEFAULT_IN()
    %  double = NLPSOL_DEFAULT_IN(int ind)
    %
    %Default input for an NLP solver.
    %
    %Extra doc: https://github.com/casadi/casadi/wiki/L_1t4
    %
    %Doc source: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/nlpsol.hpp#L222
    %
    %Implementation: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/nlpsol.cpp#L222-L226
    %
    %
    %
    %.......
    %
    %::
    %
    %  NLPSOL_DEFAULT_IN()
    %
    %
    %
    %[INTERNAL] 
    %Default input for an NLP solver.
    %
    %Extra doc: https://github.com/casadi/casadi/wiki/L_1t4
    %
    %Doc source: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/nlpsol.hpp#L222
    %
    %Implementation: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/nlpsol.cpp#L222-L226
    %
    %
    %
    %.............
    %
    %
    %.......
    %
    %::
    %
    %  NLPSOL_DEFAULT_IN(int ind)
    %
    %
    %
    %[INTERNAL] 
    %Default input for an NLP solver.
    %
    %Extra doc: https://github.com/casadi/casadi/wiki/L_1t4
    %
    %Doc source: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/nlpsol.hpp#L209
    %
    %Implementation: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/nlpsol.cpp#L209-L220
    %
    %
    %
    %.............
    %
    %
  [varargout{1:nargout}] = casadiMEX(833, varargin{:});
end
