function varargout = conic_debug(varargin)
    %CONIC_DEBUG [INTERNAL] 
    %
    %  CONIC_DEBUG(Function f, std::ostream & file)
    %  CONIC_DEBUG(Function f, char filename)
    %
    %Generate native code in the interfaced language for debugging
    %
    %Doc source: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/conic.hpp#L54
    %
    %Implementation: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/conic.cpp#L54-L58
    %
    %
    %
    %.......
    %
    %::
    %
    %  CONIC_DEBUG(Function f, std::ostream & file)
    %
    %
    %
    %[INTERNAL] 
    %Generate native code in the interfaced language for debugging
    %
    %Doc source: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/conic.hpp#L54
    %
    %Implementation: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/conic.cpp#L54-L58
    %
    %
    %
    %.............
    %
    %
    %.......
    %
    %::
    %
    %  CONIC_DEBUG(Function f, char filename)
    %
    %
    %
    %[INTERNAL] 
    %Generate native code in the interfaced language for debugging
    %
    %Doc source: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/conic.hpp#L48
    %
    %Implementation: 
    %https://github.com/casadi/casadi/blob/develop/casadi/core/conic.cpp#L48-L52
    %
    %
    %
    %.............
    %
    %
  [varargout{1:nargout}] = casadiMEX(827, varargin{:});
end
